/*
 * node_fragment.c
 *
 * Copyright 2022 Jonathan Schöbel <jonathan@Ubermos-2019>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "macro.h"
#include "log.h"
#include "status.h"

#include "data.h"
#include "text.h"
#include "validator.h"

#include "fragment.h"

#include "node_fragment.h"


#include "fragment_class.c"

struct SH_NodeFragment
{
	struct SH_Fragment base;

	/*@only@*/ char * tag;

	size_t child_n;
	/*@only@*/ struct SH_Fragment ** childs;
};

#define OPEN_TAG_BEGIN "<"
#define OPEN_TAG_END ">"
#define CLOSE_TAG_BEGIN "</"
#define CLOSE_TAG_END ">"

#define NEWLINE "\n"


static const struct fragment_methods methods = {
	(struct SH_Fragment * (*)(const struct SH_Fragment *,
	                          /*@out@*/ /*@null@*/
	                          struct SH_Status *))
	SH_NodeFragment_deepcopy,

	(void (*)(/*@only@*/ struct SH_Fragment *))
	SH_NodeFragment_free,

	(struct SH_Text * (*)(const struct SH_Fragment *,
	                      enum HTML_MODE,
	                      unsigned int, unsigned int, const char *,
	                      /*@out@*/ /*@null@*/ struct SH_Status *))
	SH_NodeFragment_to_html
};


/*@null@*/
/*@only@*/
struct SH_Fragment /*@alt struct SH_NodeFragment@*/ *
SH_NodeFragment_new (const char * tag,
                     /*@dependent@*/ SH_Data * data,
                     /*@out@*/ /*@null@*/ struct SH_Status * status)
	/*@globals NODE,
	           fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_NodeFragment * fragment;

	fragment = malloc (sizeof (struct SH_NodeFragment));
	if (fragment == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "SH_NodeFragment failed.\n");

		return NULL;
	}

	init_fragment (&(fragment->base), NODE, &methods, data, NULL);

	fragment->tag = strdup (tag);
	if (fragment->tag == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "fragment tag failed.\n");

		free (fragment);

		return NULL;
	}


	fragment->child_n = 0;
	fragment->childs = malloc (0);

	set_success (status);

	return (struct SH_Fragment *) fragment;
}

/*@null@*/
/*@only@*/
struct SH_Fragment /*@alt struct SH_NodeFragment@*/ *
SH_NodeFragment_raw_new (/*@only@*/ char * tag,
                         /*@dependent@*/ SH_Data * data,
                         /*@out@*/ /*@null@*/ struct SH_Status * status)
	/*@globals NODE,
	           fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_NodeFragment * fragment;

	fragment = malloc (sizeof (struct SH_NodeFragment));
	if (fragment == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "SH_NodeFragment failed.\n");

		return NULL;
	}

	init_fragment (&(fragment->base), NODE, &methods, data, NULL);

	fragment->tag = tag;

	fragment->child_n = 0;
	fragment->childs = malloc (0);

	set_success (status);

	return (struct SH_Fragment *) fragment;
}

void
SH_NodeFragment_free (/*@only@*/ struct SH_NodeFragment * fragment)
	/*@modifies fragment@*/
	/*@releases fragment@*/
{
	size_t index;

	free (fragment->tag);

	for (index = 0; index < fragment->child_n; index++)
	{
		SH_Fragment_free (fragment->childs[index]);
	}
	free (fragment->childs);

	free_fragment (&fragment->base);

	free (fragment);

	return;
}

/*@null@*/
/*@only@*/
struct SH_Fragment /*@alt struct SH_NodeFragment@*/ *
SH_NodeFragment_copy (const struct SH_NodeFragment * fragment,
                      /*@out@*/ /*@null@*/ struct SH_Status * status)
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_NodeFragment * copy;

	copy = malloc (sizeof (struct SH_NodeFragment));
	if (copy == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "SH_NodeFragment failed.\n");

		return NULL;
	}

	copy_fragment (&(copy->base), &(fragment->base));

	copy->tag = strdup (fragment->tag);
	if (copy->tag == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "fragment tag failed.\n");

		free (copy);

		return NULL;
	}


	copy->child_n = 0;
	copy->childs = malloc (0);

	set_success (status);

	return (struct SH_Fragment *) copy;
}

/*@null@*/
/*@only@*/
struct SH_Fragment /*@alt struct SH_NodeFragment@*/ *
SH_NodeFragment_deepcopy (const struct SH_NodeFragment * fragment,
                          /*@out@*/ /*@null@*/
                          struct SH_Status * status)
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_NodeFragment * copy;
	struct SH_Fragment * child;
	size_t index;

	copy = malloc (sizeof (struct SH_NodeFragment));
	if (copy == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "SH_NodeFragment failed.\n");

		return NULL;
	}

	copy_fragment (&(copy->base), &(fragment->base));

	copy->tag = strdup (fragment->tag);
	if (copy->tag == NULL)
	{
		set_status (status, E_ALLOC, 3,
		            "Memory allocation for "
		            "fragment tag failed.\n");

		free (copy);

		return NULL;
	}


	copy->child_n = fragment->child_n;
	copy->childs = malloc (sizeof (struct SH_NodeFragment *)
	                       * fragment->child_n);

	if (copy->child_n != 0 && copy->childs == NULL)
	{
		set_status (status, E_ALLOC, 5,
		            "Memory allocation for "
		            "fragment child failed.\n");

		free (copy->tag);
/* dangerous call to silence splint, should never be executed. */
#ifdef S_SPLINT_S
		free (copy->childs);
#endif
		free (copy);

		return NULL;
	}

	for (index = 0; index < fragment->child_n; index++)
	{
		child = SH_Fragment_copy (fragment->childs[index],
		                          status);

		if (child == NULL)
		{
			copy->child_n = index;
			SH_NodeFragment_free (copy);

			return NULL;
		}

		copy->childs[index] = child;
		child->parent = copy;
	}

	set_success (status);

	return (struct SH_Fragment *) copy;
}

bool
SH_Fragment_is_NodeFragment (const struct SH_Fragment * fragment)
	/*@*/
{
	return get_type (fragment) == NODE;
}

static inline
/*@observer@*/
char *
get_tag (const struct SH_NodeFragment * fragment)
	/*@*/
{
	return fragment->tag;
}

/*@null@*/
/*@only@*/
char *
SH_NodeFragment_get_tag (const struct SH_NodeFragment * fragment,
                         /*@out@*/ /*@null@*/ struct SH_Status * status)
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	char * tag;

	tag = strdup (get_tag (fragment));

	if (tag == NULL)
	{
		set_status (status, E_ALLOC, 4,
		            "Memory allocation for "
		            "fragment tag failed.\n");

		return NULL;
	}

	set_success (status);

	return tag;
}

/*@observer@*/
char *
SH_NodeFragment_raw_get_tag (const struct SH_NodeFragment * fragment)
{
	return get_tag (fragment);
}

/*@null@*/
/*@dependent@*/
struct SH_NodeFragment *
SH_Fragment_get_parent (const struct SH_Fragment * fragment)
	/*@*/
{
	return get_parent (fragment);
}

/*@null@*/
/*@observer@*/
struct SH_Fragment *
SH_NodeFragment_get_child (const struct SH_NodeFragment * fragment,
                           size_t index,
                           /*@out@*/ /*@null@*/
                           struct SH_Status * status)
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	if (index >= fragment->child_n)
	{
		set_status (status, E_VALUE, 2,
		            "Fragment: Child index out of range.\n");

		return NULL;
	}

	set_success (status);

	return fragment->childs[index];
}

bool
SH_NodeFragment_is_child (const struct SH_NodeFragment * fragment,
                          const struct SH_Fragment * child)
	/*@*/
{
	size_t index;

	for (index = 0; index < fragment->child_n; index++)
	{
		if (fragment->childs[index] == child)
		{
			return TRUE;
		}
	}
	return FALSE;
}

bool
SH_NodeFragment_is_descendant (const struct SH_NodeFragment * fragment,
                          const struct SH_Fragment * child)
	/*@*/
{
	size_t index;

	for (index = 0; index < fragment->child_n; index++)
	{
		if (fragment->childs[index] == child
		|| (SH_Fragment_is_NodeFragment (child)
		    && SH_NodeFragment_is_descendant (
		                        (struct SH_NodeFragment *)
		                        fragment->childs[index],
		                        child)))
		{
			return TRUE;
		}
	}
	return FALSE;
}

bool
SH_NodeFragment_append_child (struct SH_NodeFragment * fragment,
                              /*@only@*/ struct SH_Fragment * child,
                              /*@out@*/ /*@null@*/
                              struct SH_Status * status)
	/*@modifies fragment->childs@*/
	/*@modifies fragment->child_n@*/
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_Fragment ** new_childs;

	new_childs = realloc (fragment->childs,
	                      sizeof (struct SH_Fragment *)
	                      * (fragment->child_n + 1));

	if (new_childs == NULL)
	{
		set_status (status, E_ALLOC, 6,
		            "Memory allocation for "
		            "fragment child failed.\n");

/* bad code to silence splint, should never be executed. */
#ifdef S_SPLINT_S
				fragment->childs = (void *) 0x12345;
#endif
		return FALSE;
	}

	new_childs[fragment->child_n] = child;

	fragment->childs = new_childs;
	fragment->child_n++;

	set_success (status);

	return TRUE;
}

/*@null@*/
/*@only@*/
struct SH_Text *
SH_NodeFragment_to_html (const struct SH_NodeFragment * fragment,
                         enum HTML_MODE mode,
                         unsigned int indent_base,
                         unsigned int indent_step,
                         const char * indent_char,
                         /*@out@*/ /*@null@*/ struct SH_Status * status)
	/*@globals fileSystem@*/
	/*@modifies fileSystem@*/
	/*@modifies status@*/
{
	struct SH_Text * html;
	struct SH_Text * child;
	struct SH_Text * indent_text;
	size_t index;

	html = SH_Text_new (status);
	if (html == NULL)
	{
		return NULL;
	}

	if (mode == WRAP)
	{
		indent_text = SH_Text_new (status);
		if (indent_text == NULL)
		{
			SH_Text_free (html);
			return NULL;
		}

		for (index = 0; ((unsigned int) index) < indent_base;
		     index++)
		{
			if (!SH_Text_append_string (indent_text,
			                            indent_char,
			                            status))
			{
				SH_Text_free (html);
				SH_Text_free (indent_text);
				return NULL;
			}
		}

		if (!SH_Text_append_text (html, indent_text, status))
		{
			SH_Text_free (html);
			SH_Text_free (indent_text);
			return NULL;
		}
	}

	if (!SH_Text_append_string (html, OPEN_TAG_BEGIN, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, fragment->tag, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, OPEN_TAG_END, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_string (html, NEWLINE, status))
		{
			SH_Text_free (html);
			SH_Text_free (indent_text);
			return NULL;
		}
	}

	for (index = 0; index < fragment->child_n; index++)
	{
		child = SH_Fragment_to_html (fragment->childs[index],
					     mode,
					     indent_base + indent_step,
					     indent_step,
					     indent_char,
					     status);

		if (child == NULL)
		{
			SH_Text_free (html);

			if (mode == WRAP)
			{
				SH_Text_free (indent_text);
			}

			return NULL;
		}

		SH_Text_join (html, child);
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_text (html, indent_text, status))
		{
			SH_Text_free (html);
			SH_Text_free (indent_text);
			return NULL;
		}
	}

	if (!SH_Text_append_string (html, CLOSE_TAG_BEGIN, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, fragment->tag, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, CLOSE_TAG_END, status))
	{
		SH_Text_free (html);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text);
		}

		return NULL;
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_string (html, NEWLINE, status))
		{
			SH_Text_free (html);
			SH_Text_free (indent_text);
			return NULL;
		}

		SH_Text_free (indent_text);
	}

	set_success (status);

	return html;
}
